from .domain_types import ConvexPolyhedraAssembly
from .OptimalTransport import OptimalTransport
from .PowerDiagram import PowerDiagram
import numpy as np
import importlib

class TransportMap:
    def __init__(self, positions, masses=None, initial_weights=None, domain=None) -> None:
        if domain is None:
            domain = ConvexPolyhedraAssembly()
            if self.positions.shape[ 1 ] == 2:
                domain.add_box([0, 0], [1, 1])
            elif self.positions.shape[ 1 ] == 3:
                domain.add_box([0, 0, 0], [1, 1, 1])
            else:
                raise "TODO"
            
        if self.masses is None:
            N = self.positions.shape[ 1 ]
            self.masses = np.ones(N) / N

        self.pd = PowerDiagram(positions=positions, weights=initial_weights, domain=domain)
        self.masses = masses

        self.obj_max_dw = 1e-6
        self.max_iter = 100

    def optimize_weights(self) -> None:
        linear_solver = self._get_linear_solver()
        old_weights = np.copy(self.pd.weights)
        relax = 1
        for _ in range(self.max_iter):
            # derivatives
            mvs = self.pd.der_integrals_wrt_weights(stop_if_void=True)
            if mvs.error:
                ratio = 0.5
                self.pd.set_weights((1 - ratio) * old_weights + ratio * self.pd.weights)
                if self.verbosity > 1:
                    print("bim (going back)")
                continue
            old_weights = self.pd.weights

            #
            mvs.m_values[0] *= 2
            mvs.v_values -= self.masses

            A = linear_solver.create_matrix(
                self.pd.weights.shape[0],
                mvs.m_offsets,
                mvs.m_columns,
                mvs.m_values
            )

            b = linear_solver.create_vector(
                mvs.v_values
            )

            x = linear_solver.solve(A, b)

            # update weights. Attention aux poids négatifs
            self.pd.set_weights(self.pd.get_weights() - relax * x)

            nx = np.max(np.abs(x))
            if self.verbosity > 1:
                print("max dw:", nx)
            if nx < self.obj_max_dw:
                break
                
        return False
    

    def _get_linear_solver(self):
        if (self._linear_solver_inst is not None):
            return self._linear_solver_inst
        
        sparse_linear_solvers = ('CuPyx', 'Petsc', 'Scipy')
        msg = 'Available solvers are: {}.'.format(', '.join(sparse_linear_solvers))
        assert self.linear_solver in sparse_linear_solvers, msg

        for solver in (self.linear_solver,)+sparse_linear_solvers:
            try:
                module = importlib.import_module('pysdot.solvers.{}'.format(solver))
                break
            except ImportError:
                continue
        else:
            msg='Could not import any of the solver modules.'
            raise ImportError(msg)

        if self.verbosity > 0:
            print('Sucessfully imported sparse linear solver {}.'.format(solver))

        self._linear_solver_inst = module.Solver()
        return self._linear_solver_inst


def find_transport_map(positions, masses=None, initial_weights=None, domain=None):
    res = TransportMap(positions, masses, initial_weights, domain)
    res.optimize_weights()
    return res
